from slinn import utils


class Filter:
	def __init__(self, filter: str, methods: list[str]=None):
		self.filter = filter
		self.methods = ['GET', 'POST'] if methods is None else methods
		
	def check(self, text: str, method: str) -> bool:
		return utils.restartswith(text, self.filter) and method.upper() in self.methods


from slinn import Server, Filter


class Dispatcher:
	def __init__(self, *hosts: list):
		self.handles = []
		self.hosts = hosts if list(hosts) != [] else ['.*']
		
	def route(self, filter: Filter):
		def decorator(func):
			self.handles.append(Server.Handle(filter, func))
		return decorator	


from slinn import Request, Address, utils
import socket, ssl, threading, time, traceback, json, os


class Server:
	class Handle:
		def __init__(self, filter, function):
			self.filter = filter
			self.function = function
	
	def __init__(self, *dispatchers, ssl_fullchain: str=None, ssl_key: str=None, http_ver: str='2.0'):
		self.dispatchers = dispatchers
		self.server_socket = None
		self.ssl = ssl_fullchain is not None and ssl_key is not None
		self.ssl_cert, self.ssl_key = ssl_fullchain, ssl_key
		self.ssl_context = None
		self.http_ver = http_ver
		self.waiting = False
		
	def listen(self, address: Address):		
		self.host, self.port = address.host, address.port
		if socket.has_dualstack_ipv6() and '.' not in self.host and ':' not in self.host:
			self.server_socket = socket.create_server((self.host, self.port), family=socket.AF_INET6, dualstack_ipv6=True)
		elif ':' in self.host:
			self.server_socket = socket.create_server((self.host, self.port), family=socket.AF_INET6)
		else:
			self.server_socket = socket.create_server((self.host, self.port), family=(socket.AF_INET if '.' in self.host else socket.AF_INET6))
		self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		if self.ssl:
			self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
			self.ssl_context.load_cert_chain(certfile=self.ssl_cert, keyfile=self.ssl_key)
		self.server_socket.listen()
		print(f'HTTP{"S" if self.ssl else ""} server is available on http{"s" if self.ssl else ""}://{"" if "." in self.host else "["}{self.host}{"" if "." in self.host else "]"}{(":"+str(self.port) if self.port != 443 else "") if self.ssl else (":"+str(self.port )if self.port != 80 else "")}/')
		try:
			while True:
				if not self.waiting:
					threading.Thread(target=self.handle_request).start()
					time.sleep(0.15)
		except KeyboardInterrupt:
			print('Got KeyboardInterrupt, halting the application...')
			os._exit(0)
						
	def handle_request(self):
		try:
			self.waiting = True
			client_socket, client_address = self.server_socket.accept()
			if self.ssl:
				client_socket = self.ssl_context.wrap_socket(client_socket, server_side=True)
			self.waiting = False
			try:
				request = Request(client_socket.recv(8192).decode(), client_address)
				print(f'{request.method} request {request.full_link} from {request.ip}:{request.port} on {request.host}')
			except UnicodeDecodeError:
				return
			for dispatcher in self.dispatchers:
				if True in [utils.restartswith(request.host, host) for host in dispatcher.hosts]:
					for handle in dispatcher.handles:
						if handle.filter.check(request.link, request.method):
							client_socket.sendall(handle.function(request).make(request.version))
							client_socket.close()
							return
		except Exception:
			print('During handling request, an exception has occured:')
			traceback.print_exc()
			self.waiting = False


import socket


class Address:
	def __init__(self, port: int, host: str=None):
		self.port = port
		if host not in [None, '']:
			family, socktype, proto, cannonname, sockaddr = socket.getaddrinfo(host, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM, 0, socket.AI_PASSIVE)[0]
			self.host = sockaddr[0]
		else:
			self.host = ''


from slinn.address import Address
from slinn.filter import Filter
from slinn.request import Request
from slinn.http_response import HttpResponse
from slinn.http_redirect import HttpRedirect
from slinn.server import Server
from slinn.dispatcher import Dispatcher

version = 'Slinn v1.1.7 120124a from https://pypi.org/project/slinn/1.1.7'


class Request:	
	def __init__(self, http_data: str, client_address: tuple[str, int]):
		def parse_http(http):
			result = {'data': {}}
			for line in http.split('\n'):
				if len(line.split(':')) > 1:
					key, value = line.split(':')[0], ':'.join(line.split(':')[1:])
					result['data'][key.strip()]=value.strip().replace('\r','')
				elif len(line.split(' ')) == 3:
					params = line.strip().replace('\r','').split(' ')
					result = {'method': params[0], 'link': params[1], 'ver': params[2]} | result
			return result
			
		def get_args(text):
				return {} if text == '' else {pair.split('=')[0]: pair.split('=')[1] for pair in text.split('&')}
		
		self.data = {'data': {'user-agent': '', 'Accept': '', 'Accept-Encoding': '', 'Accept-Language': ''}}
		data = parse_http(http_data)
		self.data['data'].update(data['data'])
		del data['data']
		self.data.update(data)
		self.ip, self.port = client_address[:2]
		self.method = self.data['method']
		self.version = self.data["ver"]
		self.full_link = self.data["link"]
		self.host = self.data["data"]["Host"]
		self.user_agent = self.data["data"]["User-Agent"] if 'User-Agent' in self.data['data'].keys() else self.data['data']['user-agent']
		self.accept = self.data["data"]["Accept"].split(',')
		self.encoding = self.data["data"]["Accept-Encoding"].split(',')
		self.language = self.data["data"]["Accept-Language"].split(',')
		self.link = self.full_link[:(self.full_link.index('?') if '?' in self.full_link else None)]
		self.args = get_args(self.full_link[(self.full_link.index('?')+1 if '?' in self.full_link else len(self.full_link)):])
		self.cookies = {c.split('=')[0]:c.split('=')[1] for c in self.data["data"]["Cookie"].split(';')} if 'Cookie' in self.data['data'] else []


class HttpResponse:
	def __init__(self, payload, data: list[tuple]=None, status: str='200 OK', content_type: str='text/plain'):
		self.payload = str(payload).encode() if type(payload) in [str, int, float, bool, dict] else bytes(payload)
		self.data = data if data is not None else [('Content-Type', content_type), ('Server', 'Slinn')] + ([] if type(payload) in [str, int, float, bool, dict] else [('Content-Length', len(self.payload))])
		self.status = status
	
	def set_cookie(self, key: str, value):
		self.data.append(('Set-Cookie', f'{key}={value}'))
	
	def make(self, type: str='HTTP/2.0') -> str:
		return (f'{type} {self.status}'+'\r\n'+"\r\n".join([str(dat[0])+": "+str(dat[1]) for dat in self.data])+'\r\n\r\n').encode()+self.payload


import sys, os, shutil, json, base64
from slinn import version as slinn_version

RED = '\u001b[31m'
GREEN = '\u001b[32m'
BLUE = '\u001b[34m'
RESET = '\u001b[0m'
BOLD = '\u001b[1m'
GRAY = '\u001b[38;2;127;127;127m'


def arg_parse(arg: str):
	return base64.urlsafe_b64decode(arg.removeprefix('b64@').encode()+b'==').decode() if arg.startswith('b64@') else arg

def splits(string: str, delimeters: list[str]=[' ', '\n'], quotes: list[str]=[]):
		result = ['']
		current_quote = ''
		for char in str(string):
			if char in quotes:
				if current_quote == '':
					current_quote = char
					continue
				elif current_quote == char:
					current_quote = ''
					continue
			if current_quote == '':
				if char in delimeters:
					result.append('')
				else:
					result[-1] += char
			else:
				result[-1] += char
		return result

def get_args(expecting, text):
	text = text.strip()
	if text == '':
		return {}
	args = {'not_used': []}
	Ds = ['\n', ' ']
	Qs = ['"', "'", '`']
	spls  = splits(text, Ds, Qs)
	i = 0
	while i < len(spls):
		arg = str(spls[i])
		try:
			if arg.strip().endswith('='):
				W = arg.strip().removesuffix('=').strip()
				expecting.pop(expecting.index(W))
				args[W] = arg_parse(str(splits[i + 1]))
				i += 2
				continue
			if len(splits(arg, ['='], Qs)) == 2:
				spl = splits(arg, ['='])
				W = str(spl[0])
				if W not in args.keys():
					args[W] = arg_parse(str(spl[1]))
					expecting.pop(expecting.index(W))
				else:
					args[W] = [args[W]]
					args[W].append(arg_parse(str(spl[1])))
				i += 1
				continue
		except ValueError:
			pass

		if len(expecting) == 0:
			args['not_used'].append(arg_parse(arg))
			i += 1
			continue

		E = expecting.pop(0)
		args[E] = arg_parse(arg)
		i += 1
	return args

def replace_all(text: str, sss: list[str] | str, ss2: str) -> str:
	for ss1 in sss:
		text = text.replace(ss1, ss2)
	return text

def add_quotes_to_list(lst: list[str]):
		for l in lst:
			yield f'\'{l}\''

def update():
	project = open('project.json', 'r')
	project_json = json.loads(project.read())
	project.close()
	if 'apps' not in project_json.keys():
		return print(f'Updated project.json')
	apps = []
	for app in project_json['apps']:
		if os.path.isdir(app):
			apps.append(app)
	project_json['apps'] = apps
	project = open('project.json', 'w')
	project.write(json.dumps(project_json, indent=4))
	project.close()
	return print(f'Updated project.json')

def main():
	if len(sys.argv) > 1:
		if sys.argv[1].lower() == 'run':
			from slinn import Server, Address


			def config():
				file = open('project.json')
				cfg = json.loads(file.read())
				file.close()
				return cfg


			def load_imports(apps):
				imports = []
				for app in apps:
					imports.append(f'from {app} import dp_{app}')
				return imports


			def get_dispatchers(apps):
				dispachers = []
				for app in apps:
					dispachers.append(f'dp_{app}')
				return dispachers


			cfg = config()
			apps = cfg['apps'] if 'apps' in cfg.keys() else []
			port = cfg['port'] if 'port' in cfg.keys() else 8080
			host = cfg['host'] if 'host' in cfg.keys() else ''
			ssl_fullchain, ssl_key = None, None
			if 'ssl' in cfg.keys() and 'fullchain' in cfg.keys() and 'key' in cfg.keys():
				ssl_fullchain = '"'+cfg['ssl']['fullchain']+'"' if cfg['ssl']['fullchain'] else None    
				ssl_key = '"'+cfg['ssl']['key']+'"' if cfg['ssl']['key'] else None
			start = ';'.join(load_imports(apps))+f';Server({",".join(get_dispatchers(apps))}, ssl_fullchain={ssl_fullchain}, ssl_key={ssl_key}).listen(Address({port}, "{host}"))'
			print('Starting server...')
			exec(start)
		elif sys.argv[1].lower() == 'create':
			args = get_args(['name', 'host'], ' '.join(sys.argv[2:]))
			if 'name' not in args.keys():
				return print(f'{RED}The app`s name is not specified{RESET}')
			ensure_appname = replace_all(args['name'], '-&$#!@%^().,', '_')
			if os.path.isdir(ensure_appname):
				return print(f'{BLUE}The app named {args["name"]} exists{RESET}')
			os.mkdir(ensure_appname)
			with open(f'{ensure_appname}/__init__.py', 'w') as f:
				data = """from %appname%.app import dp_%appname%
""".replace('%appname%', ensure_appname)
				f.write(data)
			with open(f'{ensure_appname}/app.py', 'w') as f:
				data = """from slinn import Dispatcher, Filter, HttpResponse
 
dp_%appname% = Dispatcher(%host%)

# Write your code down here                         
""".replace('%appname%', ensure_appname).replace('%host%', '' if 'host' not in args.keys() else ', '.join(add_quotes_to_list(args['host'] if type(args['host']) == list else [args['host']])))
				f.write(data)
			fr = open('project.json', 'r')
			fj = json.loads(fr.read())
			fr.close()
			if 'apps' in fj.keys():
				fj['apps'].append(ensure_appname)
			else:
				fj['apps'] = []
			fw = open('project.json', 'w')
			fw.write(json.dumps(fj, indent=4))
			fw.close()
			update()
			return print(f'{GREEN}App successfully created{RESET}')
		elif sys.argv[1].lower() == 'delete':
			args = get_args(['name'], ' '.join(sys.argv[2:]))
			if 'name' not in args.keys():
				return print(f'{RED}The app`s name is not specified{RESET}')
			ensure_appname = replace_all(args['name'], '-&$#!@%^().,', '_')
			if not os.path.isdir(ensure_appname):
				return print(f'{BLUE}The app named {args["name"]} does not exist{RESET}')
			if input(f'{RESET}Are you sure? (y/N) >>> ').lower() not in ['y', 'yes']:
				return print(f'{RESET}Aborted')
			shutil.rmtree(ensure_appname)
			update()
			return print(f'{GREEN}App successfuly deleted{RESET}')
		elif sys.argv[1].lower() == 'update':
			return update()
		elif sys.argv[1].lower() == 'help':
			return print("""%BOLD%Slinn manager help page

Commands%RESET%:
	%cmd% run                                                              %GRAY%# Starts server
	%cmd% create {app`s name} host=(host1) host=(host2)...                 %GRAY%# Creates a new app
		Example: %cmd% create localhost host=localhost host=127.0.0.1%RESET%
	%cmd% delete {app`s name}                                              %GRAY%# Deletes an app
		Example: %cmd% delete localhost%RESET%
	%cmd% update                                                           %GRAY%# Updates project.py%RESET%
	%cmd% help                                                             %GRAY%# Prints this help%RESET%
	%cmd% version                                                          %GRAY%# Prints version of Slinn%RESET%
""".replace('%cmd%', f'py {sys.argv[0]}').replace('%GRAY%', GRAY).replace('%RESET%', RESET).replace('%BOLD%', BOLD))
		elif sys.argv[1].lower() == 'version':
			return print(slinn_version)
		else:
			return print(f'{RED}Command is not exists{RESET}')

			
if __name__ == '__main__':
	main()


import sys, venv
from slinn.default import *

if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1].lower() == 'create':
            args = get_args(['path'], ' '.join(sys.argv[2:]))
            apppath = (args['path']+'?').replace('/?', '').replace('?', '') if 'path' in args.keys() else '.'
            import os, slinn, shutil, platform
            modulepath = os.path.abspath(slinn.__file__).replace('__init__.py', '')
            if not os.path.isdir(apppath):
                os.mkdir(apppath)
            else:
                print(f'Warning: {apppath} exists')
            shutil.copyfile(modulepath+'default/manage.py', f'{apppath}/manage.py')
            shutil.copyfile(modulepath+'default/project.json.py', f'{apppath}/project.json')
            venv.create(f'{apppath}/venv')
            packages_dir = ''
            if platform.system() == 'Windows':
                packages_dir = f'{apppath}/venv/Lib/site-packages'
            else:
                packages_dir = f'{apppath}/venv/lib/python{".".join(sys.version.split(" ")[0].split(".")[:-1])}/site-packages'
            try:
                os.makedirs(packages_dir)
            except FileExistsError:
                pass
            shutil.copytree(modulepath, packages_dir+'/slinn')
            print('Project created')


from setuptools import setup

def readme():
    with open('README.md', 'r') as f:
        return f.read()

setup(name='slinn',
      version='1.1.7',
      description='HTTPS and HTTP server framework',
      packages=['slinn', 'slinn.default'],
      author='Mark Radin',
      author_email='mrybs2@gmail.com',
      url='https://slinn.miotp.ru/',
      long_description=readme(),
      long_description_content_type='text/markdown',
      python_requires='>=3.11',
      zip_safe=True)